////////////////////////////////////////////////////////////////////////////////
// main.cpp
// Vyroba vezi typu P
// Soutez v programovani, Ustredni kolo, Janske Lazne 2014
// main executable file
// (c) Jan Horacek, soutezni cislo 22
// jan.horacek@seznam.cz
// 21.06.2014
// v1.0
// license: Apache license v2.0
////////////////////////////////////////////////////////////////////////////////

#include <iostream>
#include <fstream>
#include "main.h"

using namespace std;

///////////////////////////////////////////////////////////////////////////////

// reset struktur, kam ukaldame data, pozuiva se pred nactenim dat
void reset()
{
    for (int i = 0; i < kostky_cnt; i++)
        kostky[i].cntA = kostky[i].cntR = kostky[i].cntG = kostky[i].cntB = kostky[i].cntY = 0;
}

///////////////////////////////////////////////////////////////////////////////

// nacteni souboru *.in
// pozor: rozmery (napr. 1X1) MUSI byt oddelene velkym pismenem 'X', nikoliv malym
int fload(char* filename)
{
    int index, cnt;
    string tmp, color;
    ifstream ifile(filename);

    if (ifile.fail()) return 1;

    reset();    // smazani aktualnich dat

    cout << "Loaded: " << filename << endl;

    // nacteni hlavicky
    ifile >> base_size.x;
    ifile >> base_size.y;
    ifile >> tmp;

    if (tmp == "P")         typ_veze = P;
    else if (tmp == "D")    typ_veze = D;
    else if (tmp == "PBDB") typ_veze = PBDB;

    // nacitani kostek
    while (!ifile.fail())
    {
        // nacteni dat
        ifile >> color;     // barva
        ifile >> tmp;       // typ kostky (1x1, 4x1, ...)
        ifile >> cnt;       // dostupne mnozstvi

        // parsing typu kostky
        if (tmp == "2X2")                           index = 0;
        else if ((tmp == "1X4") || (tmp == "4X1"))  index = 1;
        else if ((tmp == "1X3") || (tmp == "3X1"))  index = 2;
        else if ((tmp == "1X2") || (tmp == "2X1"))  index = 3;
        else if ((tmp == "1X1"))                    index = 4;
        else index = 0;

        // parsovani barvy
        if (color == "A") kostky[index].cntA += cnt;
        else if (color == "R") kostky[index].cntR += cnt;
        else if (color == "G") kostky[index].cntG += cnt;
        else if (color == "B") kostky[index].cntB += cnt;
        else if (color == "Y") kostky[index].cntY += cnt;
    }

    ifile.close();
    return 0;
}

///////////////////////////////////////////////////////////////////////////////

// prevod barvy do stringu pro ukladani dat
string ColorToString(Color color)
{
    if (color == A) return "A";
    else if (color == R) return "R";
    else if (color == G) return "G";
    else if (color == B) return "B";
    else if (color == Y) return "Y";
    else return "";
}

///////////////////////////////////////////////////////////////////////////////

// prevod typu kostky do stringy pro ukladani dat
string TypToRozmery(int typ_index, int rotace)
{
    // [0] = 2x2
    // [1] = 1x4
    // [2] = 1x3
    // [3] = 1x2
    // [4] = 1x1

    if (typ_index == 0) return "2x2";
    else if ((typ_index == 1) && (rotace == 0)) return "4x1";
    else if ((typ_index == 1) && (rotace == 1)) return "1x4";
    else if ((typ_index == 2) && (rotace == 0)) return "3x1";
    else if ((typ_index == 2) && (rotace == 1)) return "1x3";
    else if ((typ_index == 3) && (rotace == 0)) return "2x1";
    else if ((typ_index == 3) && (rotace == 1)) return "1x2";
    else if (typ_index == 4) return "1x1";
    else return "";
}

///////////////////////////////////////////////////////////////////////////////

// Ukladani jednoho patra, vola se prubezne
int fsave(const char* filename, int patro)
{
    int j;
    ofstream ofile;

    // u prvniho patra soubor prepiseme, pak uz jen navazujeme
    if (patro == 0)
        ofile.open(filename, std::fstream::trunc);    // rewrite
    else
        ofile.open(filename, std::fstream::app);    // append

    if (ofile.fail())
        return 1;

    // pred prvnim patrem ulozime hlavicku
    if (patro == 0)
        ofile << base_size.x << " " << base_size.x << " P" << endl;

    // ukladame jednotluve kostky
    for (j = 0; j < patra[patro].kostkyCnt; j++)
    {
        ofile << patro+1 << " " << patra[patro].kostky[j].pos.x << " " << patra[patro].kostky[j].pos.y << " " <<
        ColorToString(patra[patro].kostky[j].color) << " " << TypToRozmery(patra[patro].kostky[j].typ, patra[patro].kostky[j].rotace) << endl;
    }

    // nakonec kazdeho patra pridame prazdny radek
    ofile << endl;

    ofile.close();
    return 0;
}

///////////////////////////////////////////////////////////////////////////////

// vraci pocet kostek jednoho typu vsech barev
int zjisti_pocet_kostek(int index)
{
    return kostky[index].cntA + kostky[index].cntR + kostky[index].cntG + kostky[index].cntB + kostky[index].cntY;
}

///////////////////////////////////////////////////////////////////////////////

// Zjistuje, jestli na dane misto lze polozit kostku s danym tvarem (ten je definovat indexem "typ_index" podle main.h)
// vraci rotaci, v jake by mela byt kostka polozena, nebo ohlasi, ze na ni neni misto
// navratove hodnoty:
//  0 = nemuzu polozit
//  1 = rotace 0
//  2 = rotace 1
int muzu_polozit(int patro, Point pos, int typ_index)
{
    // kontrola poctu kostek
    if (zjisti_pocet_kostek(typ_index) < 1)
        return 0;

    if (typ_index == 4)
    {
        //1x1
        if (patra[patro].map[pos.y][pos.x] == 0)
            return 1;
    }
    else if (typ_index == 0)
    {
        //2x2
        if ((pos.x > base_size.x-2) || (pos.y > base_size.y-2)) return 0;      // kontrola okraju
        if ((patra[patro].map[pos.y][pos.x] == 0) && (patra[patro].map[pos.y+1][pos.x] == 0) &&
            (patra[patro].map[pos.y][pos.x+1] == 0) && (patra[patro].map[pos.y+1][pos.x+1] == 0))
            return 1;
    }
    else if (typ_index == 3)
    {
        //1x2
        if ((pos.x <= base_size.x-2) && (pos.y <= base_size.y-2))
        {
            // obe dve rotace mozne - prioritne rovne
            if ((patra[patro].map[pos.y][pos.x] == 0) && (patra[patro].map[pos.y][pos.x+1] == 0)) return 1;
            else if ((patra[patro].map[pos.y][pos.x] == 0) && (patra[patro].map[pos.y+1][pos.x] == 0)) return 2;
        }
        else if ((pos.y <= base_size.y-2))
        {
            // morna jen rotace "dolu" - doprava precuhuje
            if ((patra[patro].map[pos.y][pos.x] == 0) && (patra[patro].map[pos.y+1][pos.x] == 0)) return 2;
        }
        else if ((pos.x <= base_size.x-2)) {
            // mozna jen rotace doprava - dolu precuhuje
            if ((patra[patro].map[pos.y][pos.x] == 0) && (patra[patro].map[pos.y][pos.x+1] == 0)) return 1;
        } else
            return 0;
    }
    else if (typ_index == 2)
    {
        //1x3
        if ((pos.x <= base_size.x-3) && (pos.y <= base_size.y-3))
        {
            // obe dve rotace mozne - prioritne rovne
            if ((patra[patro].map[pos.y][pos.x] == 0) && (patra[patro].map[pos.y][pos.x+1] == 0) && (patra[patro].map[pos.y][pos.x+2] == 0)) return 1;
            else if ((patra[patro].map[pos.y][pos.x] == 0) && (patra[patro].map[pos.y+1][pos.x] == 0) && (patra[patro].map[pos.y+2][pos.x] == 0)) return 2;
        }
        else if ((pos.y <= base_size.y-3))
        {
            // morna jen rotace "dolu" - doprava precuhuje
            if ((patra[patro].map[pos.y][pos.x] == 0) && (patra[patro].map[pos.y+1][pos.x] == 0) && (patra[patro].map[pos.y+2][pos.x] == 0)) return 2;
        }
        else if ((pos.x <= base_size.x-3)) {
            // mozna jen rotace doprava - dolu precuhuje
            if ((patra[patro].map[pos.y][pos.x] == 0) && (patra[patro].map[pos.y][pos.x+1] == 0) && (patra[patro].map[pos.y][pos.x+2] == 0)) return 1;
        } else
            return 0;
    }
    else if (typ_index == 1)
    {
        //1x4
        if ((pos.x <= base_size.x-4) && (pos.y <= base_size.y-4))
        {
            // obe dve rotace mozne - prioritne rovne
            if ((patra[patro].map[pos.y][pos.x] == 0) && (patra[patro].map[pos.y][pos.x+1] == 0) && (patra[patro].map[pos.y][pos.x+2] == 0) && (patra[patro].map[pos.y][pos.x+3] == 0)) return 1;
            else if ((patra[patro].map[pos.y][pos.x] == 0) && (patra[patro].map[pos.y+1][pos.x] == 0) && (patra[patro].map[pos.y+2][pos.x] == 0) && (patra[patro].map[pos.y+3][pos.x] == 0)) return 2;
        }
        else if ((pos.y <= base_size.y-4))
        {
            // mozna jen rotace "dolu" - doprava precuhuje
            if ((patra[patro].map[pos.y][pos.x] == 0) && (patra[patro].map[pos.y+1][pos.x] == 0) && (patra[patro].map[pos.y+2][pos.x] == 0) && (patra[patro].map[pos.y+3][pos.x] == 0)) return 2;
        }
        else if ((pos.x <= base_size.x-4)) {
            // mozna jen rotace doprava - dolu precuhuje
            if ((patra[patro].map[pos.y][pos.x] == 0) && (patra[patro].map[pos.y][pos.x+1] == 0) && (patra[patro].map[pos.y][pos.x+2] == 0) && (patra[patro].map[pos.y][pos.x+3] == 0)) return 1;
        } else
            return 0;
    }
    return 0;
}

///////////////////////////////////////////////////////////////////////////////

// vlozi kostku do mapy
// mapu pouzivame k tomu,a bychom vedeli, ketra pole jsou obsazen  ktera ne
// intenzita v budoucnu muze byt i zaporna a lze tak prochazet vice reseni - "odebirat" kostky z mapy
// zatim se jen pricita
void zabarvi_mapu(int patro, Point pos, int intenzita, int typ_index, int rotace)
{
    if (typ_index == 0)
    {
        // 2x2
        patra[patro].map[pos.y][pos.x] += intenzita;
        patra[patro].map[pos.y+1][pos.x] += intenzita;
        patra[patro].map[pos.y][pos.x+1] += intenzita;
        patra[patro].map[pos.y+1][pos.x+1] += intenzita;
    }
    else if ((typ_index == 1) && (rotace == 2))
    {
        // 1x4
        patra[patro].map[pos.y][pos.x] += intenzita;
        patra[patro].map[pos.y+1][pos.x] += intenzita;
        patra[patro].map[pos.y+2][pos.x] += intenzita;
        patra[patro].map[pos.y+3][pos.x] += intenzita;
    }
    else if ((typ_index == 1) && (rotace == 1))
    {
        // 4x1
        patra[patro].map[pos.y][pos.x] += intenzita;
        patra[patro].map[pos.y][pos.x+1] += intenzita;
        patra[patro].map[pos.y][pos.x+2] += intenzita;
        patra[patro].map[pos.y][pos.x+3] += intenzita;
    }
    else if ((typ_index == 2) && (rotace == 2))
    {
        // 1x3
        patra[patro].map[pos.y][pos.x] += intenzita;
        patra[patro].map[pos.y+1][pos.x] += intenzita;
        patra[patro].map[pos.y+2][pos.x] += intenzita;
    }
    else if ((typ_index == 2) && (rotace == 1))
    {
        // 3x1
        patra[patro].map[pos.y][pos.x] += intenzita;
        patra[patro].map[pos.y][pos.x+1] += intenzita;
        patra[patro].map[pos.y][pos.x+2] += intenzita;
    }
    else if ((typ_index == 3) && (rotace == 2))
    {
        // 1x2
        patra[patro].map[pos.y][pos.x] += intenzita;
        patra[patro].map[pos.y+1][pos.x] += intenzita;
    }
    else if ((typ_index == 3) && (rotace == 1))
    {
        // 2x1
        patra[patro].map[pos.y][pos.x] += intenzita;
        patra[patro].map[pos.y][pos.x+1] += intenzita;
    }
    else if (typ_index == 4)
    {
        // 1x1
        patra[patro].map[pos.y][pos.x] += intenzita;
    }
}

///////////////////////////////////////////////////////////////////////////////

// poklada konstku na dane misto
Point poloz_kostku(int patro, Point pos, int typ_index, int rotace)
{
    // pridame kostku do databaze
    patra[patro].kostkyCnt++;
    patra[patro].kostky[patra[patro].kostkyCnt-1].pos = pos;
    patra[patro].kostky[patra[patro].kostkyCnt-1].rotace = rotace-1;
    patra[patro].kostky[patra[patro].kostkyCnt-1].typ = typ_index;

    // vmyslime ji barvu podle toho, jake kostky jsou k dizpocici
    if (kostky[typ_index].cntA > 0) {patra[patro].kostky[patra[patro].kostkyCnt-1].color = A; kostky[typ_index].cntA--; }
    else if (kostky[typ_index].cntR > 0) { patra[patro].kostky[patra[patro].kostkyCnt-1].color = R; kostky[typ_index].cntR--; }
    else if (kostky[typ_index].cntG > 0) { patra[patro].kostky[patra[patro].kostkyCnt-1].color = G; kostky[typ_index].cntG--; }
    else if (kostky[typ_index].cntB > 0) { patra[patro].kostky[patra[patro].kostkyCnt-1].color = B; kostky[typ_index].cntB--; }
    else if (kostky[typ_index].cntY > 0) { patra[patro].kostky[patra[patro].kostkyCnt-1].color = Y; kostky[typ_index].cntY--; }

    // zaneseme ji do mapy
    zabarvi_mapu(patro, pos, 1, typ_index, rotace);

    // vratime nejblizsi inteligentni pozici pro dalsi kostku, nebo (-1, -1), pokud uz v tomto patre neni volno
    while (patra[patro].map[pos.y][pos.x] != 0)
    {
        pos.x++;
        if (pos.x > base_size.x-1)
        {
            pos.x = 0;
            pos.y++;
        }

        // kontrola konce mapy
        if (pos.y > base_size.y-1) return Point(-1, -1);
    }

    return pos;
}

///////////////////////////////////////////////////////////////////////////////

// Poklada cele patro
//  vraci 0 pokud ok, 1 pokud dosly kostky
int poloz_patro(int patro)
{
    int rotace;
    Point pos;

    pos.x = 0;
    pos.y = 0;

    // prioritne chceme polozit velke kostky
    // [0] = 2x2 - priorita 3
    // [1] = 1x4 - priorita 3
    // [2] = 1x3 - priorita 2
    // [3] = 1x2 - priorita 1
    // [4] = 1x1 - priorita 0

    // pokladame, dokud neni patro plne
    while ((pos.x >= 0) && (pos.y >= 0))
    {
        if ((zjisti_pocet_kostek(0) > 0) || (zjisti_pocet_kostek(1) > 0))
        {
            // kostky 2x2, 1x4, 4x1
            if (zjisti_pocet_kostek(0) > zjisti_pocet_kostek(1))
            {
                // pokladame index 0
                if ((rotace = muzu_polozit(patro, pos, 0)) > 0)
                {
                    pos = poloz_kostku(patro, pos, 0, rotace);
                    continue;
                }
            } else {
                // pokladame index 1
                if ((rotace = muzu_polozit(patro, pos, 1)) > 0)
                {
                    pos = poloz_kostku(patro, pos, 1, rotace);
                    continue;
                }
            }
        }

        if (zjisti_pocet_kostek(2) > 0)
        {
            // kostky 1x3, 3x1
            if ((rotace = muzu_polozit(patro, pos, 2)) > 0)
            {
                pos = poloz_kostku(patro, pos, 2, rotace);
                continue;
            }
        }

        if (zjisti_pocet_kostek(3) > 0)
        {
            // kostky 1x2, 2x1
            if ((rotace = muzu_polozit(patro, pos, 3)) > 0)
            {
                pos = poloz_kostku(patro, pos, 3, rotace);
                continue;
            }
        }

        if (zjisti_pocet_kostek(4) > 0)
        {
            // kostka 1x1
            if ((rotace = muzu_polozit(patro, pos, 4)) > 0)
            {
                pos = poloz_kostku(patro, pos, 4, rotace);
                continue;
            }
        }

        // dosly kostky
        return 1;
    }// while

    return 0;
}

///////////////////////////////////////////////////////////////////////////////

// Stavi vez - stavi jednotliva patra
void vytvor_patra(string ofilename)
{
    int ret = 0;
    for (patra_cnt = 0; ret == 0; patra_cnt++)
    {
        // info pro uzivatele
        cout << "Stavim " << patra_cnt << ". patro" << endl;
        cout << "Pred stavenim: " << zjisti_pocet_kostek(0) << " 2x2; " << zjisti_pocet_kostek(1) << " 1x4; " << zjisti_pocet_kostek(2) << " 1x3; " << zjisti_pocet_kostek(3) << " 1x2; " << zjisti_pocet_kostek(4) << " 1x1;" << endl;

        // samotne postaveni patra
        ret = poloz_patro(patra_cnt);

        // patra prubezne ukladame (pokud nahodou nedosly kostky)
        if (ret == 0)
            fsave(ofilename.c_str(), patra_cnt);
    }

    // posledni patro nedostaveno, tak ho odebereme
    patra_cnt--;
}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

// main
int main(int argc, char* argv[])
{
    // program prebira cestu k vstupnimu souboru jako parametr
    if (argc < 2)
    {
        cerr << "Zadejte cestu k vstupnimu souboru jako parametr" << endl;
        return 1;
    }

    cout << "Nacitam soubor: " << argv[1] << endl;
    fload(argv[1]);

    //cout << "Typ veze: " << typ_veze << endl;
    cout << "Zakladna: X=" <<  base_size.x << ", Y=" << base_size.y << endl;

    // vytvorime jmeno vystupniho souboru - odebereme priponu
    string fileName = argv[1];
    fileName = fileName.substr(0, fileName.rfind("."));

    // provedeme samotny algoritmus
    vytvor_patra(fileName);     // prubezne uklada patra

    cout << "Celkem " << patra_cnt << " pater" << endl;

    return 0;
}
